import time
from array import array, ArrayType
from copy import copy

import utf8, encodings

test_plan_number=21
test_plan_approval="tdickenson"

if 67305985L==array(encodings.typecode,'\001\002\003\004')[0]:
    # big endian
    byte_order = 1
else:
    # Little-endian
    byte_order = 0

class wstring:
    """wstring; Immutable wide string class.
    
    Instances also look like a sequence of ucs-4 characters.

    This class is not intended to be a general purpose wide-character string. It's
    only use is storage of wide character data in Zope REQUESTs and ODB, to convert
    the data into a variety of character encodings, and to be well behaved in dtml
    and other bits of the zope framework.
    """

    def __init__(self,data,encoding='guess'):
        """Initialize the wide string with the specified character data, using the given encoding.
        The internal representation is created lazily, as an optimisation for the common case where
        a wstring is immediately converted back into the encoding it was created from.
        """
        encoding = encoding_aliases.get(encoding,encoding)
        if encoding=='guess':
            if isinstance(data,wstring):
                # Efficient copy
                if hasattr(data,'_ucs4'):
                    self._ucs4 = data._ucs4
                else:
                    self._lazy_init = data._lazy_init
                self._cache = data._cache
                return
            else:
                encoding = 'latin1'
        try:
            fn = encoding_functions[encoding][0]
        except KeyError:
            raise encodings.WStringError('Unknown encoding',encoding)
        self._lazy_init = lambda fn=fn,data=data: fn(data)
        self._cache = { encoding: data }

    def __getattr__(self,attr):
        if '_ucs4'==attr:
            a = self._ucs4 = self._lazy_init()
            return a
        elif known_character_encoding(attr):
            return lambda fn=self.encode, encoding=attr: fn(encoding)
        else:
            raise AttributeError(attr)

    def __hash__(self):
        return hash(self._ucs4.tostring())

    def __cmp__(self,other):
        if isinstance(other, wstring):
            return cmp(self._ucs4, other._ucs4)
        else:
            return cmp(str(self), str(other))

    def __repr__(self):
        return "<wstring %s>" % repr(self._ucs4)[11:-1]

    def __str__(self):
        # If we get included in some dtml that is not wstring aware, then guess
        # which character encoding that dtml might be expecting. latin-1 is the
        # only option.
        return self.encode('latin1')

    def __call__(self):
        # Argh - many bits of Zope assume that an attribute can be either a string,
        # or a method that returns a string when called. This is close enough
        return self

    def __getstate__(self):
        return byte_order,self._ucs4.tostring()

    def __setstate__(self,state):
        old_byte_order,s = state
        self._ucs4 = array(encodings.typecode,s)
        if old_byte_order != byte_order:
            self._ucs4.byteswap()
        self._cache = {}

    # Instances behave like immutable sequences of ucs-4 characters

    def __getitem__(self,i):
        return self._ucs4[i]

    def __getslice__(self,i,j):
        return self.__class__(self._ucs4[i:j],'ucs4')

    def encode(self,encoding):
        """Convert the string into a specified character encoding. The __getattr__ hook
        also provides method-like access to each encoding.
        """
        encoding = encoding_aliases.get(encoding,encoding)
        e = self._cache.get(encoding,None)
        if e is None:
            try:
                fn = encoding_functions[encoding][1]
            except KeyError:
                raise encodings.WStringError('Unknown encoding',encoding)
            e = self._cache[encoding] = fn(self._ucs4)
        return e




def latin1_to_ucs4(s):
    return array('B',str(s)).tolist()

def ucs4_to_latin1(ucs4):
    o = ''
    for c in ucs4:
        if c>255:
            o = o+'?'
        else:
            o = o+chr(c)
    return o

def ucs4_to_ucs4(s):
    if type(s)==ArrayType:
        return s
    elif isinstance(s,wstring):
        return wstring._ucs4
    else:
        if type(s)!=type([]):
            # Ack - array uses the list accessor functions, not the sequence protocol.
            s = list(s)
        return array(encodings.typecode,s)

encoding_aliases = { 'UTF-8':           'utf8',
                     'utf-8':           'utf8',
                     'ISO-8859-1':      'latin1',
                     'ISO_8859-1':      'latin1',
                     'ISO_8859-1:1987': 'latin1',
                   }

# A mapping from encoding id to a pair of functions.
#
# The first function converts from encoded data to ucs4, where ucs4 must
# be an instance of array.array, using the typecode from the encodings
# module. The second must be an inverse.
#
# The id needs to be usable as a ZPublisher marshalling tag, and therefore
# must be alphanumeric. More descriptive aliases can be provided in the table
# above
#
# A __getattr__ makes these ids available as method-like attributes of wstring
# instances, for example mywstring.utf8() will return the right thing.

encoding_functions = { 'utf8':   (utf8.utf8_to_ucs4, utf8.ucs4_to_utf8),
                       'latin1': (latin1_to_ucs4,    ucs4_to_latin1),
                       'ucs4':   (ucs4_to_ucs4,      ucs4_to_ucs4),
                     }


def known_character_encoding(encoding):
    return encoding_functions.has_key(encoding_aliases.get(encoding,encoding))


################################################################################

class CheckError(Exception):
    pass

class InequalityError(Exception):
    pass
 
def check(t):
    if not t: raise CheckError()

def check_equality(a,*bs):
    check(len(bs))
    for b in bs:
        if a!=b: raise InequalityError(a,b)

def test():
    tester = wstring(range(0,255)+range(0xff,0xfffff,1000)+range(0xfffff,0xfffffff,100000),'ucs4')
    check_equality(wstring(tester.utf8(),'utf8'),tester)
    hangugo = wstring([0xD55C, 0xAD6D, 0xC5B4],'ucs4')
    check_equality(hangugo.utf8(),'\355\225\234\352\265\255\354\226\264')
    check_equality(wstring(hangugo.utf8(),'utf8'),hangugo)
    check_equality(hangugo.utf8(),copy(hangugo).utf8())
    check_equality(hangugo[0],0xD55C)
    check_equality(hangugo[1],0xAD6D)
    check_equality(hangugo[2],0xC5B4)
    nihongo = wstring([0x65E5, 0x672C, 0x8A9E],'ucs4')
    check_equality(wstring(nihongo.utf8(),'utf8'),nihongo)
    check_equality(nihongo.latin1(),'???')
    check_equality('hello world',wstring('hello world').latin1())
    check_equality('hello world',wstring('hello world').utf8())


if __name__=='__main__':
    print "Test-Plan-Number: %s"%test_plan_number
    print "Test-Plan-Approval: %s"%test_plan_approval
    print "Date-Performed: %s"% time.ctime(time.time())
    test()
    print "Status: PASS\n"
